from ngsolve import *
#from netgen.geom2d import SplineGeometry
import ngsolve as ng
from ngsolve import x, y, z, curl, Mesh, HCurl, BlockMatrix
from netgen.csg import OrthoBrick, unit_cube, CSGeometry, Pnt
from math import pi, log
from array import array
import math as m
import sympy as sm
import scipy.sparse as sp
import numpy as np
from ngsolve.la import EigenValues_Preconditioner
import time

box = OrthoBrick(Pnt(-1,-1,-1),Pnt(1,1,1)).bc("outer")
geo = CSGeometry()
geo.Add(box)

beta = 10**(2)  

X, Y, Z = sm.symbols('X Y Z')

# Smooth solution
ex = ((1 - X**2)*(1 - Y**2)*(1 - Z**2),
      (1 - X**2)*(1 - Y**2)*(1 - Z**2),
      (1 - X**2)*(1 - Y**2)*(1 - Z**2))


def smcurl(M):
    return (sm.diff(M[2], Y) - sm.diff(M[1], Z),
            sm.diff(M[0], Z) - sm.diff(M[2], X),
            sm.diff(M[1], X) - sm.diff(M[0], Y))


def str2coef(symbol):
    print(symbol)
    return ng.CoefficientFunction(eval(str(symbol)
                                       .replace("X", "x")
                                       .replace("Y", "y")
                                       .replace("Z", "z")))

exact = str2coef(ex)
curlex = smcurl(ex)
curl_exact = str2coef(curlex)
curlcurl_exact = str2coef(smcurl(smcurl(ex)))
src = curlcurl_exact -beta*exact

P_order = 1
Ref_num = 5
Int_mesh_size = 2
        
ngmesh = geo.GenerateMesh(maxh=Int_mesh_size)
mesh = Mesh(ngmesh)
Draw(exact,mesh,"exact")

def setup_space(mesh):
    V = HCurl(mesh, order=P_order, type1=True,dirichlet="outer" )
    W = HCurl(mesh, order=P_order, type1=True,dirichlet="outer")
    W.SetOrder(TRIG,2)
    W.SetOrder(TET,3)
    W.Update()

    edges_no=mesh.nedge
    faces_no=mesh.nface
    elements_no=mesh.nv
    print('#ne=', mesh.ne)
    print('#nedge=', mesh.nedge)
    print('#nface=', mesh.nface)
    print('#nv=', mesh.nv)
    print ("V.ndof =", V.ndof)
    print ("W.ndof =", W.ndof)
    
    
    # define trial- and test-functions
    u,v = V.TnT()
    p,q = W.TnT()
    
    # the bilinear-form
    a = ng.BilinearForm(V,)# symmetric=True)
    a += ng.SymbolicBFI(curl(u) * curl(v)-beta*u*v)
    
    b = ng.BilinearForm(trialspace=V, testspace=W)
    b += ng.SymbolicBFI(curl(p) * curl(v)-beta*p*v)
    
    c = ng.BilinearForm(W)
    c += ng.SymbolicBFI(curl(p) * curl(q)-beta*p*q)
    
    # the right hand side
    fv = ng.LinearForm(V)
    fv += ng.SymbolicLFI(src*v, simd_evaluate=True)
    
    fw = ng.LinearForm(W)
    fw += ng.SymbolicLFI(src*q)
    
    # the solution field
    gfv = ng.GridFunction(V)
    gfw = ng.GridFunction(W)
    
    ng.Draw(gfv, mesh, "gfv")
    ng.Draw(gfw, mesh, "gfw")
    

    return [V,W,gfv, gfw, a, b, c, fv, fw]

def SolveBVP(mesh,V,W,gfv, gfw, a, b, c, fv, fw): 

    with TaskManager():
        bitarray = W.FreeDofs()
        for j in range(mesh.nedge):
            bitarray[j] = 0
        gfv.Update()
        gfw.Update()
        a.Assemble()
        b.Assemble()
        c.Assemble()
        fv.Assemble()
        fw.Assemble()
        gfv.vec.data = a.mat.Inverse(V.FreeDofs()) * fv.vec
        r = fw.vec.CreateVector()
        r.data = fw.vec - b.mat * gfv.vec
        gfw.vec.data = c.mat.Inverse(bitarray) * r

    Redraw (blocking=True)
    return [gfv,gfw]
beta1=abs(beta)
def CalcError(mesh,gfv,gfw):
    curl_u_gfv = curl(gfv)
    curl_e_gfw = curl(gfw)

    # Errors norms of exact error (u-u')
    L2_u_err = ng.sqrt(ng.Integrate((gfv-exact)*(gfv-exact), mesh))   # L2 norm of the exact error

    HCurl_u_err = ng.sqrt(ng.Integrate(
        (gfv-exact)*(gfv-exact) +
        (curl_u_gfv-curl_exact)*(curl_u_gfv-curl_exact), mesh))       # HCurl norm of the exact error

    Enr_u_err = ng.sqrt(ng.Integrate(
        beta1*(gfv-exact)*(gfv-exact) +
        (curl_u_gfv-curl_exact)*(curl_u_gfv-curl_exact), mesh))       # Energy norm of the exact error

    # Errors norms of error function (E)
    L2_e_err = ng.sqrt(ng.Integrate((gfw)*(gfw), mesh))               # L2 norm of error function

    HCurl_e_err = ng.sqrt(ng.Integrate(
        (gfw)*(gfw) +
        (curl_e_gfw)*(curl_e_gfw), mesh))                             # HCurl norm of error function

    Enr_e_err = ng.sqrt(ng.Integrate(
        beta1*(gfw)*(gfw) +
        (curl_e_gfw)*(curl_e_gfw), mesh))                             # Energy norm of error function

    L2_u_err_list.append(L2_u_err)
    HCurl_u_err_list.append(HCurl_u_err)
    Enr_u_err_list.append(Enr_u_err)
    L2_e_err_list.append(L2_e_err)
    HCurl_e_err_list.append(HCurl_e_err)
    Enr_e_err_list.append(Enr_e_err)


    #---------------------------------------------
    preI_a= a.mat.CreateSmoother(V.FreeDofs())
    ##preI_a=IdentityMatrix(V.ndof, complex=False)
    A_lams = EigenValues_Preconditioner(mat=a.mat, pre=preI_a)
    A_CN_list.append(abs(max(A_lams)/min(A_lams)))

    preI=c.mat.CreateSmoother(W.FreeDofs())
    ##preI=IdentityMatrix(W.ndof, complex=False)
    C_lams = EigenValues_Preconditioner(mat=c.mat, pre=preI)
    C_CN_list.append(max(C_lams)/min(C_lams))


def WriteVTK(mesh,gfv,exact,i):
    vtk = VTKOutput(ma=mesh,
                    coefs=[gfv,exact],
                    names = ["gfv","exact"],
                    filename="Exampl_uniform"+str(i) ,
                    subdivision=0)
    # Exporting the results:
    vtk.Do()

L2_u_err_list = array('d', [0])
HCurl_u_err_list = array('d', [0])
Enr_u_err_list = array('d', [0])
L2_e_err_list = array('d', [0])
HCurl_e_err_list = array('d', [0])
Enr_e_err_list = array('d', [0])
Mesh_size_list = array('d', [0])
V_DOF_list = array('d', [0])
A_CN_list = array('d', [0])
C_CN_list = array('d', [0])

start_time = time.time()

with TaskManager():
    for i in range(Ref_num):
        Mesh_size = Int_mesh_size/2**(i)
        Mesh_size_list.append(Mesh_size)
        if i > 0:
            ngmesh.Refine()
        mesh = Mesh(ngmesh)
        V,W,gfv, gfw, a, b, c, fv, fw = setup_space(mesh)
        V_DOF_list.append(V.ndof)
        gfv, gfw = SolveBVP(mesh,V,W,gfv, gfw, a, b, c, fv, fw)
        CalcError(mesh,gfv, gfw)
        #WriteVTK(mesh,gfv,exact,i)
        #input("yes")

print("--- %s seconds ---" % (time.time() - start_time))

#st=33
## VTKOutput object
vtk = VTKOutput(ma=mesh,
                coefs=[gfv,exact],
                names = ["gfv","exact"],
                filename="Exampl_uniform" ,
                subdivision=0)
# Exporting the results:
vtk.Do()
#=============================================================================
print("------------------------------L2 Norms --------------------------------")
print("-----------------------------------------------------------------------")
print("|", "  h  ","|", "n_dof(V)", "|","L2 Err.(U)","|","L2 Err.(E)", "|", "order" ,"|", "V_order" ,"|", "Eff." , "|")
print("-----------------------------------------------------------------------")
for i in range(1, Ref_num+1):
    if i == 1:
        print("|", "%1.3f" % Mesh_size_list[i],
        "|", " %6d " % V_DOF_list[i],
        "|", "%8.4E" % L2_u_err_list[i], "|", "%8.4E" %  L2_e_err_list[i], "|", "     ", "|", "       ", "|",
        "%3.2f" % (L2_e_err_list[i]/L2_u_err_list[i]), "|")
    else:
        rateL2 = log(L2_e_err_list[i-1]/L2_e_err_list[i])/log(2)
        rate_V = log(L2_e_err_list[i-1]/L2_e_err_list[i])/log(V_DOF_list[i-1]/V_DOF_list[i])
        print("|", "%1.3f" % Mesh_size_list[i], "|", " %6d " % V_DOF_list[i],
        "|", "%8.4E" % L2_u_err_list[i], "|", "%8.4E" %  L2_e_err_list[i], "|","%1.3f" % rateL2, "|","%1.3f " % rate_V, "|",
        "%3.2f" % (L2_e_err_list[i]/L2_u_err_list[i]), "|")
print("-----------------------------------------------------------------------")

#=============================================================================
print("-------------------------------HCurl Norms ------------------------------")
print("-------------------------------------------------------------------------")
print("|", "  h  ","|", "n_dof(V)", "|","HCurl Err_U","|","HCurl Err_E", "|", "order" ,"|", "V_order" ,"|", "Eff." , "|")
print("-------------------------------------------------------------------------")
for i in range(1, Ref_num+1):
    if i == 1:
        print("|", "%1.3f" % Mesh_size_list[i],"|", " %6d " % V_DOF_list[i],
        "|", "%8.4E " % HCurl_u_err_list[i], "|", "%8.4E "
    % HCurl_e_err_list[i], "|", "     ", "|", "       ", "|",
        "%3.2f" % (HCurl_e_err_list[i]/HCurl_u_err_list[i]), "|")
    else:
        rateHcurl = log(HCurl_e_err_list[i-1]/HCurl_e_err_list[i])/log(2)
        rate_V = log(HCurl_e_err_list[i-1]/HCurl_e_err_list[i])/log(V_DOF_list[i-1]/V_DOF_list[i])
        print("|", "%1.3f" % Mesh_size_list[i],"|", " %6d " % V_DOF_list[i],
        "|", "%8.4E " % HCurl_u_err_list[i], "|", "%8.4E "
    % HCurl_e_err_list[i], "|","%1.3f" % rateHcurl, "|","%1.3f " % rate_V, "|",
        "%3.2f" % (HCurl_e_err_list[i]/HCurl_u_err_list[i]), "|")
print("-------------------------------------------------------------------------")
#=============================================================================
print("-------------------------------Energy Norms ------------------------------")
print("-------------------------------------------------------------------------")
print("|", "  h  ","|", "n_dof(V)", "|","Enr Err_U ","|","Enr Err_E ", "|", "order" ,"|", "V_order" ,"|", "Eff." , "|")
print("-------------------------------------------------------------------------")
for i in range(1, Ref_num+1):
    if i == 1:
        print("|", "%1.3f" % Mesh_size_list[i],"|", " %6d " % V_DOF_list[i],
        "|", "%8.4E" % Enr_u_err_list[i], "|", "%8.4E"
    % Enr_e_err_list[i], "|", "     ", "|", "       ", "|",
        "%3.2f" % (Enr_e_err_list[i]/Enr_u_err_list[i]), "|")
    else:
        rateEnr = log(Enr_e_err_list[i-1]/Enr_e_err_list[i])/log(2)
        rate_V = log(Enr_e_err_list[i-1]/Enr_e_err_list[i])/log(V_DOF_list[i-1]/V_DOF_list[i])
        print("|", "%1.3f" % Mesh_size_list[i],"|", " %6d " % V_DOF_list[i],
        "|", "%8.4E" % Enr_u_err_list[i], "|", "%8.4E"
    % Enr_e_err_list[i], "|","%1.3f" % rateEnr, "|","%1.3f " % rate_V, "|",
        "%3.2f" % (Enr_e_err_list[i]/Enr_u_err_list[i]), "|")
print("-------------------------------------------------------------------------")
#=============================================================================
print("--------Condition Numbers----------")
print("-----------------------------------")
print("|", "  h  ", "|", "Cnd No (A)", "|", "Cnd NO (C)", "|")
print("-----------------------------------")
for i in range(1, Ref_num+1):
    print("|", "%1.3f" % Mesh_size_list[i],
    "|", " %08.3f " % A_CN_list[i], "|", " %08.3f " % C_CN_list[i], "|")
print("-----------------------------------")



